#!/usr/bin/env python
# $Id: queuesim.py,v 1.7 2006/04/27 04:15:17 cvs Exp $

# Simulate queuing using Markov chains

import sys,signal
from fpformat import *
from random import *

# To fix "broken pipe" weirdness 
signal.signal(signal.SIGPIPE, signal.SIG_DFL)  

# a few globals I'm afraid -- connected with model evaluation

F=[]
G=[]
NStates= 64
ModelOverflow=1000000

def printUsage():
  print >> sys.stderr,"Usage: queuesim chaintype",
  print >> sys.stderr,"pktsize bandwidth time parm1 parm2 [dt]"
  print >> sys.stderr,"       1= Wang model p1= a  p2 = alpha"
  print >> sys.stderr,"       2= PSST(a) p1=q p2 = a"
  print >> sys.stderr,"       3= Arrow/Barenco p1=aL p2=aR"
  print >> sys.stderr,"       4= Clegg/Dodson p1=mean p2= alpha"
  print >> sys.stderr,"       5= PSST(b) p1=q p2 = a"
  print >> sys.stderr,"       6= Bernouilli  p1=mean p2= unused"
  sys.exit()  

def initWang(a,alpha):
  "Initialise simulation for Wang model"
  for i in range (0, NStates):
    F.append (a /(pow((i+1.0),(1.0+alpha))))
#   print F[i]
  
  s= WangIter(a,alpha,0)
  return s[0]

def WangIter(a,alpha,state):
#Usually we just head left
  if state > 0:
    return (state-1, 1)
  # The earliest states are explicitly enumerated
  r= random()
  for i in range(0,NStates):
#    print r, F[i]
    if r > F[i]:
#      print i
      return (i,0)
#  print >> sys.stderr, "Got to the hard bit"
  minim= NStates
  while minim < ModelOverflow:  # Arbitrary stopping point of 100000
    maxim= minim*2
    if (inWangInfRange(minim,maxim,minim,alpha)):
      while minim != maxim:
        newmaxim= int((minim+maxim)/2)
        if (inWangRange(minim, newmaxim, minim, maxim,alpha)):
          maxim= newmaxim
        else:
          minim= newmaxim+1
      return (minim, 0)
    minim= maxim+1
  print >> sys.stderr,"Overflow of state in Wang at state",ModelOverflow
  return (ModelOverflow,0)

def inWangInfRange (i,j,k,alpha):
  r=random()
  f= pow(i,-(alpha+1.0)) - pow(j+1.0,-(alpha+1.0))
  f/= pow (k,-(alpha+1.0))
  if r < f:
    return 1
  return 0

def inWangRange (i,j,k,l,alpha):
  r= random()
  f= pow(i,-(alpha+1.0)) - pow(j+1.0,-(alpha+1.0))
  f/= (pow (k,-(alpha+1.0)) - pow(l+1.0,-(alpha+1.0)))
  if r < f:
    return 1
  return 0

def initPSST(q,a):
  "Initialise simulation state for PSST model"
  for i in range (0, NStates):
    F.append (pow(a,-(i+1.0))*a /(a-1.0) )
#    print >> sys.stderr, F[i]
  (state,out)= PSSTIter(q,a,0)
  return state
  
def PSSTIter(q,a,state,rev= 0):
  "Iteration of PSST model -- parameter rev reverses ones and zeros"
  if state > 0:
    for i in range(0,state):
      r= random()
      if r > float(q)/a:
        return (state,rev)
    return(0,rev)
  r= random()
  for i in range(0, NStates):
    if r > F[i]:
      return (i, 1-rev)
  minim= NStates
  while minim < ModelOverflow:  # Arbitrary stopping point of 100000
    maxim= minim*2
    if (inPSSTInfRange(minim,maxim,minim,a)):
      while minim != maxim:
        newmaxim= int((minim+maxim)/2)
        if (inPSSTRange(minim, newmaxim, minim, maxim,a)):
          maxim= newmaxim
        else:
          minim= newmaxim+1
      return (minim, 1-rev)
    minim= maxim+1
  print >> sys.stderr,"Overflow of state in PSST at state",ModelOverflow
  return (ModelOverflow,1-rev)

def inPSSTInfRange(i,j,k,a):
  "Prob X in [i,j] | X >= k"
  f= pow(a,-i) - pow(a,-(j+1.0));
  f/= pow(a,-k)
  r= random()
  if r < f:
    return 1
  return 0

def inPSSTRange(i,j,k,l,a):
  "Prob X in [i,j] | X in [k,l]"
  f= pow(a,-i) - pow(a,-(j+1.0));
  f/= (pow(a,-k) - pow(a, - (l+1.0)));
  r= random()
  if r < f:
    return 1
  return 0

def initArrow(alphal,alphar):
  "Initialise simulation state for Arrowsmith/Barenco model"
  for i in range (0, NStates):
    F.append (1 /(pow((i+1.0),(1.0+alphal))))
    G.append (1 /(pow((i+1.0),(1.0+alphar))))
  if random() < 0.5:
    state= 1
  else:
    state= -1
  (state,out)= ArrowIter(alphal, alphar, state)
  return state

def ArrowIter (alphal, alphar, state):
  "Perform iteration of Arrowsmith/Barenco model"
  if state < -1:
    return (state + 1, 0)
  if state > 1:
    return (state -1, 1)
  if state == 1:
    alpha= alphal
    retval= 1
    statemult= -1
  else:
    alpha= alphar
    retval= 0
    statemult= 1
  r= random()
  for i in range(1,NStates):
    if state == 1 and r > F[i]:
      return (-i,1)
    elif state == -1 and r > G[i]:
      return (i,0)
  minim= NStates
  while minim < ModelOverflow:  # Arbitrary stopping point of 100000
    maxim= minim*2
    if (inWangInfRange(minim,maxim,minim,alpha)):
      while minim != maxim:
        newmaxim= int((minim+maxim)/2)
        if (inWangRange(minim, newmaxim, minim, maxim,alpha)):
          maxim= newmaxim
        else:
          minim= newmaxim+1
      return (minim*statemult, retval)
    minim= maxim+1
  print >> sys.stderr,"Overflow of state in Arrow at state",ModelOverflow
  return (ModelOverflow*statemult,retval)
  

def initClegg(mean,alpha):
  "Initialise simulation state for Clegg/Dodson model"
  for i in range (0, NStates):
    F.append((mean/(1.0 - mean)) *(pow((i+1.0),-alpha) - \
      pow( (i+2.0), -alpha)))
      
  (state,out)= CleggIter(mean,alpha,0)
#    for i in range(0,NStates):
#      print >> sys.stderr, F[i]
  return state
  
def CleggIter(mean,alpha,state):
  "Perform iteration of Clegg/Dodson model"
  #Usually we just head left
  if state > 0:
    return (state-1, 1)
  # The earliest states are explicitly enumerated
  r= random()
  for i in range(0,NStates):
    if r > F[i]:
      return (i,0)
#  print >> sys.stderr, "Got to the hard bit"
  minim= NStates
  while minim < ModelOverflow:  # Arbitrary stopping point of 100000
    maxim= minim*2
    if (inInfRange(minim,maxim,minim,alpha)):
      while minim != maxim:
        newmaxim= int((minim+maxim)/2)
        if (inRange(minim, newmaxim, minim, maxim,alpha)):
          maxim= newmaxim
        else:
          minim= newmaxim+1
      return (minim, 0)
    minim= maxim+1
  print >> sys.stderr,"Overflow of state in Clegg/Dodson"
  return (ModelOverflow,0)

def inRange (i,j,k,l,alpha):
  r= random()
  f= pow(i,-alpha) - pow(i+1.0,-alpha) - pow (j+1.0, -alpha) + \
    pow (j+2.0, -alpha)
  f/= (pow(k,-alpha) - pow (k+1.0,-alpha) - pow (l+1.0, -alpha) + \
    pow (l+2.0, -alpha))
  if r < f:
    return 1
  return 0
  
def inInfRange (i,j,k,alpha):
  r= random()
  f= pow(i,-alpha) - pow(i+1.0,-alpha) - pow (j+1.0, -alpha) + \
    pow (j+2.0, -alpha)
  f/= (pow (k,-alpha) - pow(k+1.0, -alpha));
  if r < f:
    return 1
  return 0

# Main starts here

narg= len(sys.argv)
dt= 0
if narg != 7 and narg != 8:
  printUsage()
try:
  model= int(sys.argv[1])
  pktsize= int(sys.argv[2])
  bw= int(sys.argv[3])
  endtime = float (sys.argv[4])
  p1= float(sys.argv[5])
  p2= float(sys.argv[6])
  if narg == 8:
    dt= float(sys.argv[7])
except ValueError:
  printUsage()
seed()
if dt == 0:
  dt= float(pktsize)/bw
time= 0
if model == 1:
  state= initWang(p1,p2)
elif model == 2:
  state= initPSST(p1,p2)
elif model == 3:
  state= initArrow(p1,p2)
elif model == 4:
  state= initClegg(p1,p2)
elif model == 5:
  state= initPSST(p1,p2)
elif model == 6:
  state= 0
else:
  print >> sys.stderr, "No such model as type",model
  printUsage()
print >> sys.stderr, "Sim: pktsz:",pktsize,"b/w:",bw,"Time:",endtime, \
  "dt:",dt 
#binout= tcpdumpread.TCPDumpRead()
lastpkttime= -1
pkts= 0
nullpkts= 0
while time < endtime:
  if model == 1:
    (state,out)= WangIter(p1,p2,state)
  elif model == 2:
    (state,out)= PSSTIter(p1,p2,state) 
  elif model == 3:
    (state,out)= ArrowIter(p1,p2,state)
  elif model == 4:
    (state,out)= CleggIter(p1,p2,state)
  elif model == 5:
    (state,out)= PSSTIter(p1,p2,state,1)
  elif model == 6:
    if random() < p1:
      out= 1
    else:
      out= 0
  else:
    print >> sys.stderr,"Model not yet implemented",model
    sys.exit()
   
  if out == 1:
    if lastpkttime == -1:
      print 0,pktsize,0
    else:
      print time-lastpkttime,pktsize,0
    lastpkttime= time
    pkts+=1
  else:
    nullpkts+=1
  time+=dt
print >> sys.stderr,"Packets:",pkts,"Gaps:",nullpkts,"Ratio:", \
  float(pkts)/(pkts+nullpkts),"Total:",pkts+nullpkts
#binout.printit()
