#!/usr/bin/env python
# $Id: queueup.py,v 1.5 2006/04/27 19:27:51 cvs Exp $

# Output distribution of packet transmission rates

import tcpdumpread,sys,signal
from fpformat import *

signal.signal(signal.SIGPIPE, signal.SIG_DFL)  # To fix "broken pipe" weirdness 

def printUsage():
  print >> sys.stderr, "Usage: [-b# -q] queueup.py inputfile outputfile rate"
  print >> sys.stderr, "    extra-output options: -q= queue pdf (pkts)"
  print >> sys.stderr, "                        : -b= queue pdf (bits)"
  print >> sys.stderr, "                        : -s= queue summary only"
  print >> sys.stderr, "                        : -t= queue as time series"
  print >> sys.stderr, "                        : -o#= output as on/off series"
  print >> sys.stderr, "                             # is the packet length"
  sys.exit()  

def printBinary(outpkts, pktsize, rate):
  time= 0
  pkt= 0
  qsize= 0
  dt= pktsize/rate
  print >> sys.stderr,"Binary approximation: pkt size",pktsize,
  print >> sys.stderr,"rate:",rate,"bits/t dt:",dt
  nextarr=outpkts[0][0]
  binout= tcpdumpread.TCPDumpRead()
  pktsent= 0
  lastpkttime= -1
  while nextarr >= 0 or qsize >= pktsize:
    # if appropriate send a packet
    
    # if a packet is to arrive it should be added to queue
    if nextarr < time + dt and nextarr != -1:
      qsize+=outpkts[pkt][1]
      pkt+=1
      if pkt < len(outpkts):
        nextarr+=outpkts[pkt][0]
      else:
        nextarr= -1  # flags that we have run out of packets
      continue
    time+= dt 
    if qsize >= pktsize:
      qsize-=pktsize
      if lastpkttime == -1:
        binout.append(0,pktsize,0)
        lastpkttime= time
      else:
        binout.append(time-lastpkttime,pktsize,0)
        lastpkttime= time
  binout.printit()
  

def printQocc(queueTimes,rate,occ,summary=0,type="packets"):
  "get PDF for packets or bits"
  qtype= 0
  if type == "packets":
    qtype= 2
  elif type == "bits":
    qtype= 1
  if qtype == 0:
    print >> sys.stderr, "printQocc unknown type",repr(type)
    sys.exit()
  time= 0
  pdf={}
  val= 0
  for i in queueTimes:
    delta=i[0]-time
    if pdf.has_key(val):
      pdf[val]+= delta
    else:
      pdf[val]= delta
    time=i[0]
    val=i[qtype]
  sum= 0
  qsum= 0.0
  qsqsum= 0.0
  for i in sorted(pdf):
    qsum+= pdf[i]*i/time
    qsqsum+= pdf[i]*i*i/time
#    print >> sys.stderr, "i",i,"Pdf",pdf[i]/time,"E[q]=",qsum,"E[q^2]",qsqsum
    sum+=pdf[i]
    if summary == 0:
      print i,float(sum)/time
  if summary == 0:
    print >> sys.stderr, "E[q]=",qsum,"E[q^2]",qsqsum
  else:
    print >> sys.stderr, rate, occ, qsum, qsqsum

# Main starts here

narg= len(sys.argv)
if narg < 4:
  printUsage()
arg= 1
bapp= 0  # If > 0 output binary approx
qocc= 0
bocc= 0
qtime= 0
summary=0
while arg <= narg:
  if sys.argv[arg][0] != '-':
    break
  elif len(sys.argv[arg]) == 1:
    printUsage()
  elif sys.argv[arg][1] == 't':
    qtime= 1
  elif sys.argv[arg][1] == 'q':
    qocc= 1
  elif sys.argv[arg][1] == 'b':
    bocc= 1
  elif sys.argv[arg][1] == 's':
    summary = 1
  elif sys.argv[arg][1] == 'o':
    if len(sys.argv[arg]) == 2:
      printUsage()
    try:
      bapp= int(sys.argv[arg][2:])
    except ValueError:
      printUsage()
  else:
    print >> sys.stderr, "Unknown option"+sys.argv[arg]
    printUsage()
  arg= arg+1
if arg > narg:
  printUsage()
infile= sys.argv[arg]
arg= arg+1
if arg > narg:
  printUsage()
outfile= sys.argv[arg]
arg= arg+1
if arg >= narg:
  printUsage()
try:
  rate= float(sys.argv[arg])
except ValueError:
  print >> sys.stderr, "Got \"",sys.argv[arg],"\" expecting queue rate."
  printUsage()
if rate <= 0: 
  print >> sys.stderr, "Rate must be positive"
  printUsage()
inpkts= tcpdumpread.TCPDumpRead(infile)
outpkts= tcpdumpread.TCPDumpRead()
queue=[]
#Accounting info, time, change in time, packets and bits
pktct= 0  #packet number next to arrive
queuetime=[(0,0,0)]
time= 0
tdelt= 0
inPkts= 0
outPkts= 0
noBits= 0
qBits= 0
lastDepTime= -1;
nextQueueDep= -1;
nextQueueArr= inpkts[pktct][0]
nextQueueLen= inpkts[pktct][1]
#Do the queuing
while 1:
  if nextQueueDep == -1 and nextQueueArr == -1:
    break
  if nextQueueDep == -1 or (nextQueueArr < nextQueueDep and nextQueueArr != -1):
     #packet joins queue
    tdelt= nextQueueArr
    time= time+tdelt
    inPkts= inPkts+1
    qBits= qBits+nextQueueLen
    queue.append(nextQueueLen)
    # Update when next packet should leave queue
    if nextQueueDep != -1:
      nextQueueDep-= tdelt
    else:
      nextQueueDep= (queue[0])/rate
    # Update next arrival
    pktct= pktct+1
    if pktct < len(inpkts):
      nextQueueArr= inpkts[pktct][0]
      nextQueueLen= inpkts[pktct][1]
    else:
      nextQueueArr= -1
  else:
    #packet leaves queue
    tdelt=nextQueueDep
    time= time+tdelt
    outPkts= outPkts+1
    qBits=qBits-queue[0]
    noBits+=queue[0]
    if lastDepTime == -1:
      outpkts.append(0,queue[0],0)
    else:
      outpkts.append(time-lastDepTime,queue[0],0)
    lastDepTime= time
      #remove one from queue
    if len(queue) == 1:
      queue = []
    else:
      queue= queue[1:]
      #Update when next packet should depart
    if nextQueueArr != -1:
      nextQueueArr-= tdelt
      # Get next departure if there is one
    if len(queue) > 0:
      nextQueueDep= queue[0]/rate
    else:
      nextQueueDep= -1  
    # Record the state
#    print time,qBits,len(queue)
  queuetime.append((time,qBits,len(queue)))
if summary == 0:
    print >> sys.stderr, "In p %5.3g"%inPkts,"Out p %5.3g"%outPkts,"Bits %5.3g"%noBits,
    print >> sys.stderr, "Time %5.3g"%time,"Rate %5.3g"%(float(noBits)/time), 
    print >> sys.stderr, 'Occ %5.3g'% (float(noBits)/(rate*time))
outpkts.write(outfile) 

if qtime == 1:
  for i in queuetime:
    print i[0],i[1],i[2]
if qocc == 1:
  printQocc(queuetime,rate,(float(noBits)/(rate*time)),summary,"packets")
if bocc == 1:
  printQocc(queuetime,rate,(float(noBits)/(rate*time)),summary,"bits")
if bapp > 0:
  printBinary(outpkts,bapp,rate)
